// Copyright (c) 2014, Google Inc.
// SPDX-License-Identifier: ISC

/* This header is provided in order to make compiling against code that expects
   OpenSSL easier. */

#include "obj.h"
#include "asn1.h"
